<?php

namespace Drupal\microspid\Service;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Component\Serialization\Json;
use Psr\Log\LoggerInterface;

/**
 * Service to interact with the Idps list.
 */
class IdpsManager {

  protected $config;

  protected $queryURL;

  protected $idps;

  protected $idpsDev;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * {@inheritdoc}
   */
  public function __construct(ConfigFactoryInterface $config_factory, LoggerInterface $logger) {
    $this->config = $config_factory->get('microspid.settings');
    $this->logger = $logger;
    $this->queryURL = "https://registry.spid.gov.it/entities-idp?&output=json&custom=info_display_base";
    if ($this->getIdpsList()) {
      $this->idps = Json::decode($this->getIdpsList());
    }
    else {
      $this->idps = Json::decode('
        [
          {
            "entity_id": "https://id.eht.eu",
            "organization_name": "EtnaHitech S.C.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-etnaid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fid.eht.eu?output=json"
          },
          {
            "entity_id": "https://id.lepida.it/idp/shibboleth",
            "organization_name": "Lepida S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-lepidaid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fid.lepida.it%2Fidp%2Fshibboleth?output=json"
          },
          {
            "entity_id": "https://identity.infocert.it",
            "organization_name": "InfoCert S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-infocertid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fidentity.infocert.it?output=json"
          },
          {
            "entity_id": "https://identity.sieltecloud.it",
            "organization_name": "Sielte S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-sielteid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fidentity.sieltecloud.it?output=json"
          },
          {
            "entity_id": "https://idp.namirialtsp.com/idp",
            "organization_name": "Namirial",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-namirialid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fidp.namirialtsp.com%2Fidp?output=json"
          },
          {
            "entity_id": "https://login.id.tim.it/affwebservices/public/saml2sso",
            "organization_name": "TI Trust Technologies srl",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-timid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Flogin.id.tim.it%2Faffwebservices%2Fpublic%2Fsaml2sso?output=json"
          },
          {
            "entity_id": "https://loginspid.aruba.it",
            "organization_name": "ArubaPEC S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-arubaid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Floginspid.aruba.it?output=json"
          },
          {
            "entity_id": "https://posteid.poste.it",
            "organization_name": "Poste Italiane SpA",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-posteid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fposteid.poste.it?output=json"
          },
          {
            "entity_id": "https://spid.intesa.it",
            "organization_name": "IN.TE.S.A. S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-intesaid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fspid.intesa.it?output=json"
          },
          {
            "entity_id": "https://spid.register.it",
            "organization_name": "Register.it S.p.A.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-spiditalia.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fspid.register.it?output=json"
          },
          {
            "entity_id": "https://spid.teamsystem.com/idp",
            "organization_name": "TeamSystem s.p.a.",
            "logo_uri": "https://cdn.statically.io/gh/italia/spid-sp-access-button/master/src/production/img/spid-idp-teamsystemid.svg",
            "registry_link": "https://registry.spid.gov.it/entities-idp/https%3A%2F%2Fspid.teamsystem.com%2Fidp?output=json"
          }
        ]
      ');
    }

    $this->idpsDev = Json::decode('
      [
        {
          "organization_name": "Test ID",
          "entity_id": "spid-testenv-identityserver",
          "logo_uri": "img/idp-logo-testid.svg"
        },
        {
          "organization_name": "DemoIPD",
          "entity_id": "http://localhost:8088",
          "logo_uri": "img/idp-logo-demo.svg"
        },
        {
          "organization_name": "TestOnLine ID",
          "entity_id": "https://idptest.spid.gov.it",
          "logo_uri": "img/spid-ico-circle-bb.svg"
        }
      ]
    ');
  }

  /**
   * Return idps basic data.
   *
   * @return array
   */
  public function getIdps(): array {
    $debug = $this->config->get('test_mode');
    if ($this->config->get('test_mode')) {
      return $this->idpsDev;
    }
    return $this->idps;
  }

  /**
   * Download IDP metadata from Agid server
   */
  public function metadataDownload(): bool {
    $file = \Drupal::service('file_system')->realpath('private://microspid') . '/spid-entities-idps.xml';
    $bak = $file . '.bak';
    $url = 'https://registry.spid.gov.it/metadata/idp/spid-entities-idps.xml';
    if (file_exists($file) && !@copy($file, $bak)) {
      return FALSE;
    }
    $success = copy($url, $file);
    // debug
    if ($this->config->get('debug')) {
      $this->logger->debug(t('copy result: ' . $success));
    }
    if (!$success && !file_exists($file)) {
      rename($bak, $file);
    }
    @unlink($bak);
    return $success;
  }

  public function getIdpsList() {
    try {
      $response = \Drupal::httpClient()
        ->get($this->queryURL, [
          // 'auth' => ['username', 'password'],
        ]);
    } catch (\Throwable $th) {
      //throw $th;
      $this->logger->error(t($th->__toString()));
      return false;
    }
    // debug
    if ($this->config->get('debug')) {
      $this->logger->debug(t('getIdpsList result: ' . $response->getBody()));
    }

    return (string) $response->getBody();
  }
}
